library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity cameracontrol is
  Port (
    --Camera interface
    DATA : in STD_LOGIC_VECTOR (9 downto 0);
    CAMERA_CLK : out STD_LOGIC;
    RESET_N : out STD_LOGIC;
    LINE_VALID : in STD_LOGIC;
    FRAME_VALID : in STD_LOGIC;
    SPI_ENABLE : out STD_LOGIC;
    SPI_CLK : out STD_LOGIC;
    SPI_DATA : out STD_LOGIC;

    --Valid signal for 80/20 MHz clocks
    clk_locked : in std_logic;
    
    --Reset input, active low
    reset_in : in std_logic;

    --Camera read signals
    camera_enable : in std_logic;
    clk_80mhz : in std_logic;
    mem_write_en : out std_logic_vector(0 downto 0);
    mem_write_addr : out std_logic_vector(16 downto 0);
    mem_write_d : out std_logic_vector (31 downto 0);
    current_line_addr : out std_logic_vector(16 downto 0);
    line_complete : out std_logic;
    frame_complete : out std_logic;
    
    --SPI configuration signals
    clk_20mhz : in std_logic;
    spi_write_enable : in std_logic;
    spi_busy : out std_logic;
    spi_address : in std_logic_vector(3 downto 0);
    spi_data_in : in std_logic_vector(11 downto 0)
  );
end cameracontrol;

architecture Behavioral of cameracontrol is
  signal pixel_valid : std_logic;
  signal reset_int : std_logic;
  signal spi_dataword : std_logic_vector(15 downto 0);
  signal spi_busy_i : std_logic;
begin
  --Keep internal reset asserted until reset input is deasserted 
  --and 80/20 MHz clocks are valid. Assert reset when SPI is active.
  reset_int <= '1' when reset_in = '1' and clk_locked = '1' and spi_busy_i = '0' else '0';

  RESET_N <= reset_int;
  CAMERA_CLK <= clk_80mhz;
  pixel_valid <= LINE_VALID and FRAME_VALID;
  spi_busy <= spi_busy_i;

-------------------------------------------
--READ FROM CAMERA AND WRITE DATA TO MEMORY
-------------------------------------------

cam_read : process(clk_80mhz, FRAME_VALID, pixel_valid, reset_int, DATA, camera_enable)
  type pixelarray is array (2 downto 0) of std_logic_vector(9 downto 0);
  variable pixels : pixelarray;
  variable idx : integer range 0 to 2;
  variable address : unsigned(16 downto 0);
  variable line_address : unsigned(16 downto 0);
  variable mem_write_addr_i : unsigned(16 downto 0);
  variable valid : boolean;
  variable px_data_i : std_logic_vector(29 downto 0);
  
  type state_type is (idle, readpx, writepx);
  variable state : state_type;
  variable new_state : state_type;
begin
  if reset_int = '0' then
    valid := false;
    address := (others=>'0');
    line_address := (others=>'0');
    idx := 0;
    mem_write_en <= "0";
    pixels(2) := (others=>'0');
    pixels(1) := (others=>'0');
    pixels(0) := (others=>'0');
  	px_data_i := (others=>'0');
    state := idle;
    new_state := idle;
  elsif rising_edge(clk_80mhz) then
    state := new_state;
  
    case state is
    when readpx =>
      if pixel_valid = '1' then pixels(idx) := DATA; end if;
      if idx = 2 or pixel_valid = '0' then
        new_state := writepx;
      else
        idx := idx + 1;
      end if;
    when writepx =>
      idx := 1;
      mem_write_addr_i := address;
      address := address+1;
      px_data_i := pixels(2) & pixels(1) & pixels(0);
      pixels(2) := (others=>'0');
      pixels(1) := (others=>'0');
      pixels(0) := (others=>'0');
      if pixel_valid = '1' then
        pixels(0) := DATA;
        new_state := readpx;
      else
        new_state := idle;
      end if;
    when others => --idle or undefined state
      if camera_enable = '1' and pixel_valid = '1' then
	    idx := 0;
        new_state := readpx;
      end if;
      if FRAME_VALID = '0' then
        address := (others=>'0');
      end if;
      line_address := address;
    end case;
  end if;
  
  mem_write_d <= "00" & px_data_i;
  mem_write_addr <= std_logic_vector(mem_write_addr_i);
  current_line_addr <= std_logic_vector(line_address);
  if state = writepx then
    mem_write_en <= "1";
    if pixel_valid = '0' then
      line_complete <= '1';
    else
      line_complete <= '0';
    end if;
    if FRAME_VALID = '0' then
      frame_complete <= '1';
    else
      frame_complete <= '0';
    end if;
  else
    mem_write_en <= "0";
    line_complete <= '0';
    frame_complete <= '0';
  end if;

end process;



----------------------------------------
--SPI WRITE TO CAMERA INTERNAL REGISTERS
----------------------------------------

spi_dataword <= spi_address & spi_data_in when rising_edge(spi_write_enable);

spi_write : process (reset_int, spi_write_enable, spi_dataword, clk_20mhz)
  type state_type is (idle, spiwrite, spicomplete);
  variable state : state_type;
  variable new_state : state_type;
  variable idx : integer range 0 to 16;
begin
  if reset_int = '0' then
    state := idle;
    new_state := idle;
    idx := 16;
    SPI_DATA <= '0';
    SPI_ENABLE <= '0';
    SPI_CLK <= '0';
    spi_busy_i <= '0';
  else
    if falling_edge(clk_20mhz) then
      state := new_state;
      case state is
      when spiwrite =>
        if idx = 0 then
          new_state := spicomplete;
        else
          idx := idx - 1;
        end if;
      when spicomplete =>
        new_state := idle;
      when others =>
        idx := 16;
        if spi_write_enable = '1' then
          new_state := spiwrite;
        end if;
      end case;
    end if;
    
    case state is
    when spiwrite =>
      SPI_DATA <= spi_dataword(idx);
      SPI_ENABLE <= '0';
      SPI_CLK <= clk_20mhz;
      spi_busy_i <= '1';
    when spicomplete =>
      SPI_DATA <= '0';
      SPI_ENABLE <= '1';
      SPI_CLK <= '0';
      spi_busy_i <= '1';
    when others =>
      SPI_DATA <= '0';
      SPI_ENABLE <= '0';
      SPI_CLK <= '0';
      spi_busy_i <= '0';
    end case;
  end if;
end process;

end Behavioral;