library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use ieee.numeric_std.all;
library UNISIM;
use UNISIM.VComponents.all;

entity pcie_app is
generic(
  DATAPATH       : integer := 64;
  CORE_FREQ    : integer:=250
);
port(
  clk     : in std_logic;
  reset_n : in std_logic;
  
  int_request : out std_logic;
  int_ack : in std_logic;

  slv_dataout   : in std_logic_vector((DATAPATH-1)  downto 0);
  slv_bytevalid : in std_logic_vector((DATAPATH/8-1) downto 0);
  slv_bytecount : in std_logic_vector(12 downto 0);
  slv_addr      : in std_logic_vector(63 downto 0);
  slv_bar       : in std_logic_vector(6 downto 0);
  slv_readreq   : in std_logic;
  slv_writereq  : in std_logic;
  slv_write     : in std_logic;
  slv_lastwrite : in std_logic;
  slv_accept    : out std_logic;
  slv_abort     : out std_logic;
  slv_ur        : out std_logic;
  
  dma_rd        : in std_logic;
  dma_rdaddr    : in std_logic_vector(15 downto 0);
  dma_rdchannel : in std_logic_vector(15 downto 0);
  dma_rddata    : out std_logic_vector((DATAPATH-1) downto 0);
  dma_wr        : in std_logic;
  dma_wrchannel : in std_logic_vector(15 downto 0);
  dma_wrdata    : in std_logic_vector((DATAPATH-1) downto 0);
  
  dma0_regin    : out std_logic_vector(127 downto 0);
  dma0_regout   : in std_logic_vector(127 downto 0);
  dma0_param    : out std_logic_vector(23 downto 0);
  dma0_control  : out std_logic_vector(5 downto 0);
  dma0_status   : in std_logic_vector(3 downto 0);
  dma0_fifocnt  : out std_logic_vector(12 downto 0);
  
  CAM0_DATA        : in STD_LOGIC_VECTOR (9 downto 0);
  CAM0_CLK         : out STD_LOGIC;
  CAM0_RESET_N     : out STD_LOGIC;
  CAM0_LINE_VALID  : in STD_LOGIC;
  CAM0_FRAME_VALID : in STD_LOGIC;
  CAM0_SPI_ENABLE  : out STD_LOGIC;
  CAM0_SPI_CLK     : out STD_LOGIC;
  CAM0_SPI_DATA    : out STD_LOGIC
);
end pcie_app;

architecture pcie_app_arch of pcie_app is

component blk_mem_32_64
  port (
  clka: IN std_logic;
  wea: IN std_logic_VECTOR(0 downto 0);
  addra: IN std_logic_VECTOR(16 downto 0);
  dina: IN std_logic_VECTOR(31 downto 0);
  clkb: IN std_logic;
  addrb: IN std_logic_VECTOR(15 downto 0);
  doutb: OUT std_logic_VECTOR(63 downto 0));
end component;

component clk_80_20
port
 (-- Clock in ports
  CLK_IN1           : in     std_logic;
  -- Clock out ports
  CLK_OUT1          : out    std_logic;
  CLK_OUT2          : out    std_logic;
  LOCKED            : out    std_logic
 );
end component;

component cameracontrol
  Port (
    DATA : in STD_LOGIC_VECTOR (9 downto 0);
    CAMERA_CLK : out STD_LOGIC;
    RESET_N : out STD_LOGIC;
    LINE_VALID : in STD_LOGIC;
    FRAME_VALID : in STD_LOGIC;
    SPI_ENABLE : out STD_LOGIC;
    SPI_CLK : out STD_LOGIC;
    SPI_DATA : out STD_LOGIC;

    clk_locked : in std_logic;
    reset_in : in std_logic;

    camera_enable : in std_logic;
    clk_80mhz : in std_logic;
    mem_write_en : out std_logic_vector(0 downto 0);
    mem_write_addr : out std_logic_vector(16 downto 0);
    mem_write_d : out std_logic_vector (31 downto 0);
    current_line_addr : out std_logic_vector(16 downto 0);
    line_complete : out std_logic;
    frame_complete : out std_logic;
    
    clk_20mhz : in std_logic;
    spi_write_enable : in std_logic;
    spi_busy : out std_logic;
    spi_address : in std_logic_vector(3 downto 0);
    spi_data_in : in std_logic_vector(11 downto 0)
  );
end component;

constant host_address : std_logic_vector(63 downto 0) := (others=>'0');
constant lines_per_transmission : integer := 4;
constant bytes_per_line : integer := 856;

signal clk_80mhz : std_logic;
signal clk_20mhz : std_logic;
signal clk_locked : std_logic;
signal mem_write_en : std_logic_vector(0 downto 0);
signal mem_write_addr : std_logic_vector(16 downto 0);
signal mem_write_d : std_logic_vector (31 downto 0);
signal current_line_addr : std_logic_vector(16 downto 0);
signal line_complete : std_logic;
signal frame_complete : std_logic;
signal camera_enable : std_logic;
signal spi_write_enable : std_logic;
signal spi_busy : std_logic;
signal spi_address : std_logic_vector(3 downto 0);
signal spi_data_in : std_logic_vector(11 downto 0);

begin

--DMA Parameters (3.5) Hard code to RAM mode, 1 cycle latency, mem write burst
dma0_param <= "00" & "000" & "0000000" & "0111" & "0000" & "00" & "0" & "0";
dma0_fifocnt <= (others=>'0');--Not used in RAM mode
camera_enable <= '1';

-------------------------------------------
--Process to handle bus slave transactions
-------------------------------------------

process(clk, reset_n, slv_dataout, slv_bytevalid, slv_bytecount, slv_addr, 
  slv_bar, slv_readreq, slv_writereq, slv_write, slv_lastwrite, spi_busy)
  type state_type is (idle, accept, abort, unsupported, writebar2);
  variable state : state_type;
  variable new_state : state_type;
begin
  if reset_n = '0' then
    state := idle;
    new_state := idle;
  elsif rising_edge (clk) then
    state := new_state;
    case state is
    when idle =>
      if slv_writereq = '1' then
        if slv_bar = "0000001" then
          state := accept;
        else
          state := unsupported;
        end if;
      elsif slv_readreq = '1' then
        state := unsupported;
      end if;
    when writebar2 =>
	  --go to idle state after writing the first word
	    if slv_write = '1' then
        state := idle;
      end if;
    when accept =>
      state := writebar2;
    when others =>
      state := idle;
    end case;
  end if;
  
  
  if state = writebar2 and slv_write = '1' then
      spi_write_enable <= '1';
      spi_address <= slv_addr(6 downto 3);
      spi_data_in <= slv_dataout(11 downto 0);
  else
    spi_write_enable <= '0';
    spi_address <= (others=>'0');
    spi_data_in <= (others=>'0');
  end if;

  slv_abort <= '0';
  if state = unsupported then slv_ur <= '1'; else slv_ur <= '0'; end if;
  if state = accept then slv_accept <= '1'; else slv_accept <= '0'; end if;
    
end process;


-------------------------------------------
--Process to handle bus master transactions
-------------------------------------------

process(line_complete, clk, reset_n, current_line_addr)
  type state_type is (idle, sendline, initialize);
  variable state : state_type;
  variable new_state : state_type;
  variable cnt : integer range 0 to 480;
  variable lines_read : integer range 0 to 480;
  variable local_address : std_logic_vector(16 downto 0);
  variable dma_control_i : std_logic_vector(5 downto 0);
begin
  if reset_n = '0' then
    state := idle;
    new_state := initialize;
    cnt := 0;
    lines_read := 0;
    dma_control_i := "000000";
    dma0_control <= (others=>'0');
    dma0_regin <= (others=>'0');
  else
    if rising_edge(line_complete) then
      if cnt = 0 then
        local_address := current_line_addr;
      end if;
      cnt := cnt+1;
      if cnt = lines_per_transmission then
        lines_read := cnt;
        cnt := 0;
        new_state := sendline;
      end if;
    end if;
    
    if rising_edge(clk) then
      state := new_state;
      new_state := idle;
    
      case state is
      when sendline =>
        --Abort ongoing transfer, write enable for size and local address registers
        dma0_control <= "101100";
      when initialize =>
        --Write enable for DMA parameters, PCI address
        dma0_control <= "010011";
      when others =>
        dma0_control <= "000000";
      end case;

    end if;
    
    --Channel register (3.4), Transfer Size & Local address
    dma0_regin(95 downto 64) <= std_logic_vector(to_unsigned(lines_read*bytes_per_line, 32));
    dma0_regin(127 downto 96) <= "0000000000000000" & local_address(16 downto 1);
    --Channel register (3.4), PCI address
    dma0_regin(63 downto 0) <= std_logic_vector(unsigned(host_address) + (unsigned(local_address(16 downto 1)) & "000"));
  end if;
end process;

--------------------
-- Interrupt process
--------------------

int_process: process(reset_n, clk, frame_complete, int_ack)
  type state_type is (idle, interrupt);
  variable state : state_type;
  variable new_state : state_type;
begin
  if reset_n = '0' then
    state := idle;
    new_state := idle;
    int_request <= '0';
  else
    if rising_edge(frame_complete) then
      new_state := interrupt;
    end if;
    
    if rising_edge(clk) then
      state := new_state;
      
      if int_ack = '1' then
        new_state := idle;
      end if;
    end if;
    
    if state = interrupt then
      int_request <= '1';
    else
      int_request <= '0';
    end if;
  end if;
end process;
 
buffer_mem : blk_mem_32_64
port map (
  clka => clk_80mhz,
  wea => mem_write_en,
  addra => mem_write_addr,
  dina => mem_write_d,
  clkb => clk,
  addrb => dma_rdaddr,
  doutb => dma_rddata
);

clk_gen : clk_80_20
port map (
  CLK_IN1 => clk,
  CLK_OUT1 => clk_80mhz,
  CLK_OUT2 => clk_20mhz,
  LOCKED => clk_locked
);

cam0 : cameracontrol
port map (
  DATA => CAM0_DATA,
  CAMERA_CLK => CAM0_CLK,
  RESET_N => CAM0_RESET_N,
  LINE_VALID => CAM0_LINE_VALID,
  FRAME_VALID => CAM0_FRAME_VALID,
  SPI_ENABLE => CAM0_SPI_ENABLE,
  SPI_CLK => CAM0_SPI_CLK,
  SPI_DATA => CAM0_SPI_DATA,
  
  clk_locked => clk_locked,
  reset_in => reset_n,
  
  camera_enable => camera_enable,
  clk_80mhz => clk_80mhz,
  mem_write_en => mem_write_en,
  mem_write_addr => mem_write_addr,
  mem_write_d => mem_write_d,
  current_line_addr => current_line_addr,
  line_complete => line_complete,
  frame_complete => frame_complete,
  
  clk_20mhz => clk_20mhz,
  spi_write_enable => spi_write_enable,
  spi_busy => spi_busy,
  spi_address => spi_address,
  spi_data_in => spi_data_in
);
end pcie_app_arch;

